/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.impl;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.fastmessenger.model.IMessenger;
import com.fastmessenger.model.IReceiver;
import com.fastmessenger.model.IReturn;

/**
 * A messenger implementation
 * 
 * @author Rex Young
 */
public class Messenger implements IMessenger {

    private ExecutorService executorService;

    private AddressManager addressManager = new AddressManager ();
    private RouteManager routeManager = new RouteManager ();
    private ReceiverManager receiverManager = new ReceiverManager ();
    private ReplyManager replyManager = new ReplyManager ();

    /**
     * Constructor that will create a fixed size thread pool
     */
    public Messenger () {
        this (Executors.newFixedThreadPool (Runtime.getRuntime ().availableProcessors ()));
    }

    /**
     * Constructor that will take an outside ExecutorService
     * 
     * @param executorService
     *            outside ExecutorService
     */
    public Messenger (ExecutorService executorService) {
        this.executorService = executorService;

        registerReceiver (replyManager, replyManager.SelfAddress.getActiveobject (),
                replyManager.SelfAddress.getService ());
    }

    /**
     * Returns the ExecutorService the messenger is using
     * 
     * @return the ExecutorService the messenger is using
     */
    public ExecutorService getExecutorService () {
        return executorService;
    }

    synchronized public void registerReceiver (Object implObj, String receiver, String... services) {

        registerReceiverImpl (ReceiverType.NonGui, new Object[] { implObj }, receiver, services);
    }

    synchronized public void registerReceiver (Object[] implObjs, String receiver, String... services) {
        registerReceiverImpl (ReceiverType.NonGui, implObjs, receiver, services);
    }

    synchronized public void deregisterReceiver (String receiver) {

        Utils.chkNonEmptyId (receiver);

        if (addressManager.getSortedServices (receiver) != null) {
            addressManager.removeReceiver (receiver);
            receiverManager.removeReceiver (receiver);
        }
    }

    synchronized public void sendMessage (String receiverService, Object... serviceArguments) {

        Address serviceAddress = new Address (receiverService);

        placeServiceRequest (null, null, 0, serviceAddress, serviceArguments);
    }

    synchronized public <T> IReturn<T> callService (String receiverService, Object... serviceArguments) {

        Address serviceAddress = new Address (receiverService);

        return placeServiceRequest (replyManager.SelfAddress, null, 0, serviceAddress, serviceArguments);
    }

    synchronized public void callServiceFor (Object[] listenerInfo, String receiverService,
            Object... serviceArguments) {

        Address serviceAddress = new Address (receiverService);

        int returnValueIndex;
        Address returnAddress;
        Object[] returnArguments;

        if (listenerInfo[0] instanceof String) {
            returnValueIndex = -1;
            returnAddress = new Address ((String) listenerInfo[0]);
            returnArguments = new Object[listenerInfo.length - 1];
            System.arraycopy (listenerInfo, 1, returnArguments, 0, returnArguments.length);
        } else {
            returnValueIndex = ((Number) listenerInfo[0]).intValue ();
            if (returnValueIndex < 0 || returnValueIndex >= listenerInfo.length - 2) {
                throw new IllegalArgumentException ("AppLogic error: returnValueIndex is out of range: "
                        + returnValueIndex);
            }
            returnAddress = new Address ((String) listenerInfo[1]);
            returnArguments = new Object[listenerInfo.length - 2];
            System.arraycopy (listenerInfo, 2, returnArguments, 0, returnArguments.length);
        }

        placeServiceRequest (returnAddress, returnArguments, returnValueIndex, serviceAddress,
                serviceArguments);
    }

    synchronized public void registerRoute (String senderService, String receiverService) {

        Address senderAddress = new Address (senderService);
        Address receiverAddress = new Address (receiverService);

        routeManager.addRoute (senderAddress, receiverAddress);
    }

    synchronized public void deregisterRoute (String senderService, String receiverService) {

        Address senderAddress = new Address (senderService);
        Address receiverAddress = new Address (receiverService);

        routeManager.removeRoute (senderAddress, receiverAddress);
    }

    synchronized public void publishMessage (String senderService, Object... serviceArguments) {

        Address senderAddress = new Address (senderService);

        List<Address> serviceAddressList = routeManager.getReceiversBySender (senderAddress);

        if (serviceAddressList != null) {
            for (Address serviceAddress : serviceAddressList) {
                try {
                    placeServiceRequest (null, null, 0, serviceAddress, serviceArguments);
                } catch (RuntimeException e) {
                }
            }
        }
    }

    synchronized void registerReceiverImpl (ReceiverType type, Object[] implObjs, String receiver,
            String[] services) {

        Utils.chkNonDupArray (implObjs, false);
        Utils.chkSameTypeArray (implObjs);
        Utils.chkNonEmptyId (receiver);
        Utils.chkNonDupArray (services, true);
        if (!(implObjs[0] instanceof IReceiver)) {
            Utils.chkMethodNames (implObjs[0], services);
        }

        if (addressManager.getSortedServices (receiver) == null) {
            addressManager.addReceiver (receiver, services);
            receiverManager.addReceiver (receiver, this, type, implObjs);
        } else {
            throw new IllegalArgumentException ("AppLogic error: receiverId already exist: " + receiver);
        }
    }

    synchronized <T> IReturn<T> placeServiceRequest (Address returnAddress, Object[] returnArguments,
            int returnValueIndex, Address serviceAddress, Object[] serviceArguments) {

        String[] sortedServices;

        if (returnAddress != null && returnAddress != replyManager.SelfAddress) {

            sortedServices = addressManager.getSortedServices (returnAddress.getActiveobject ());

            if (sortedServices == null) {

                throw new IllegalArgumentException ("AppLogic error: returnId not found: " + returnAddress);

            } else if (returnAddress.getService ().isEmpty ()) {

                returnAddress = new Address (returnAddress.getActiveobject (), sortedServices[0]);

            } else if (Arrays.binarySearch (sortedServices, returnAddress.getService ()) < 0) {

                throw new IllegalArgumentException ("AppLogic error: returnService not found: "
                        + returnAddress);
            }
        }

        sortedServices = addressManager.getSortedServices (serviceAddress.getActiveobject ());

        if (sortedServices == null) {

            throw new IllegalArgumentException ("AppLogic error: receiverId not found: " + serviceAddress);

        } else if (serviceAddress.getService ().isEmpty ()) {

            serviceAddress = new Address (serviceAddress.getActiveobject (), sortedServices[0]);

        } else if (Arrays.binarySearch (sortedServices, serviceAddress.getService ()) < 0) {

            throw new IllegalArgumentException ("AppLogic error: receiverService not found: "
                    + serviceAddress);
        }

        Method serviceMethod = receiverManager.findServiceMethodIfAny (serviceAddress, serviceArguments);

        ServiceRequest request = new ServiceRequest (returnAddress, returnArguments, returnValueIndex,
                serviceAddress, serviceMethod, serviceArguments);

        if (returnAddress == null) {

            receiverManager.addServiceRequest (request);
            return null;

        } else {
            Reply<T> reply;

            if (returnAddress == replyManager.SelfAddress) {
                reply = new Reply<T> ();
                String replyId = replyManager.registerPendingReply (reply);
                reply.setReplyId (replyId);

                request.setReturnArguments (new Object[] { replyId, null });
                request.setReturnValueIndex (1);
            } else {
                reply = null;
            }

            receiverManager.addServiceRequest (request);

            return reply;
        }
    }

}
