/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.impl;

import java.lang.reflect.Method;
import java.util.LinkedList;

import com.fastmessenger.model.IReceiver;

class ReceiverMailbox {

    protected Messenger messenger;

    protected boolean mailboxRunning = true;
    protected LinkedList<ServiceRequest> requestQueue = new LinkedList<ServiceRequest> ();

    protected Head[] headcounts;

    protected Object lockObj = requestQueue;

    protected class Head {
        boolean running = false;
        Object implObj;

        Head (Object implObj) {
            this.implObj = implObj;
        }
    }

    ReceiverMailbox (Messenger messenger, Object[] implObjs) {
        this.messenger = messenger;

        headcounts = new Head[implObjs.length];
        for (int i = 0; i < implObjs.length; i++) {
            headcounts[i] = new Head (implObjs[i]);
        }
    }

    Method findServiceMethodIfAny (String service, Object... serviceArguments) {

        if (headcounts[0].implObj instanceof IReceiver) {
            return null;
        } else {
            return Utils.getMethodByValues (headcounts[0].implObj.getClass (), service, serviceArguments);
        }
    }

    void stop () {
        synchronized (lockObj) {
            mailboxRunning = false;
        }
    }

    void addServiceRequest (ServiceRequest request) {
        synchronized (lockObj) {
            if (mailboxRunning) {
                requestQueue.add (request);
                launchHeadIfIdle ();
            }
        }
    }

    private void launchHeadIfIdle () {
        for (int i = 0; i < headcounts.length; i++) {
            if (!headcounts[i].running) {

                final Head head = headcounts[i];
                messenger.getExecutorService ().submit (new Runnable () {
                    public void run () {
                        while (true) {
                            ServiceRequest request;
                            synchronized (lockObj) {
                                if (!mailboxRunning || requestQueue.isEmpty ()) {
                                    head.running = false;
                                    return;
                                }
                                request = requestQueue.poll ();
                            }
                            executeServiceRequest (head, request);
                        }
                    }
                });

                headcounts[i].running = true;
                return;
            }
        }
    }

    protected void executeServiceRequest (Head head, ServiceRequest request) {
        Object returnValue;
        try {
            if (head.implObj instanceof IReceiver) {

                returnValue = ((IReceiver) head.implObj).onServiceRequest (request.getServiceAddress ()
                        .getService (), request.getServiceArguments ());
            } else {

                returnValue = request.getServiceMethod ().invoke (head.implObj,
                        request.getServiceArguments ());
            }
        } catch (Throwable e) {
            returnValue = e;
            e.printStackTrace (System.err);
        }

        if (request.getReturnAddress () != null) {
            try {
                if (request.getReturnValueIndex () >= 0) {
                    request.getReturnArguments ()[request.getReturnValueIndex ()] = returnValue;
                }

                messenger.placeServiceRequest (null, null, 0, request.getReturnAddress (),
                        request.getReturnArguments ());

            } catch (Throwable e) {
            }
        }
    }

}
