/*
 * Copyright (C) 2012-2013 The Dojo Foundation
 * 
 * Copyright (C) 2012-2013 Rex Young
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.fastmessenger.model;

/**
 * This interface defines a messenger object that acts as a virtual machine to host and realize
 * active objects.
 * 
 * A messenger hosts active objects and manages their lifecycles by creating them, binding them with
 * implementation objects, and destroying them. These functions are supported by registerReceiver()
 * and deregisterReceiver() methods.
 * 
 * A messenger realizes active objects by providing caller objects accesses to active objects, and
 * forwarding service requests to implementation objects.
 * 
 * Accesses to active objects are supported by callService(), callServiceFor(), sendMessage(), and
 * publishMessage() methods. In addition, the publishMessage() uses dynamic routing that depends on
 * routes managed by registerRoute() and deregisterRoute() methods.
 * 
 * This interface does not define how a messenger should forward caller's calls (aka service
 * requests) to corresponding implementation objects. However the messenger implementation in this
 * Fast Messenger Programming reference implantation supports two mechanisms of forwarding.
 * 
 * First, an implementation object may choose to implement a marker interface IReceiver which
 * explicitly tells the messenger how to forward requests to the implementation object.
 * 
 * Second, a messenger forwards requests to an implementation object that is not marked by IReceiver
 * using reflection. In this case, the messenger chooses and invokes a method on the implementation
 * object based on the service name and arguments from the request.
 * 
 * A name for an individual receiver or service is consisted of letters, numbers, dots, hyphens, and
 * underlines. In order to specify a service on an active object, a qualified service name with a
 * separator of ":" like "<ActiveObjectName>:<ServiceName>" can be used. In addition,
 * "<ActiveObjectName >" without a service name can be used if the active object has only one
 * service defined. (Note: Letters may include unicodes for other languages.)
 * 
 * A service may be viewed as a method, function, or message port according to programming styles
 * user choose. A message port can produce and receive messages (namely service requests). This is
 * similar to the concept of a socket port.
 * 
 * @author Rex Young
 */
public interface IMessenger {

    /**
     * Requests to invoke a service on an active object with parameters. The service and the active
     * object are specified by the qualified service name receiverService. The parameters are
     * specified by serviceArguments which could be empty.
     * 
     * This method returns an IReturn through which the caller can either query or wait for a result
     * of the service.
     * 
     * @param receiverService
     *            qualified service name
     * @param serviceArguments
     *            arguments for the service
     * @return an instance of IReturn
     */
    <T> IReturn<T> callService (String receiverService, Object... serviceArguments);

    /**
     * Requests to invoke a service on an active object with parameters. The service and the active
     * object are specified by the qualified service name receiverService. The parameters are
     * specified by serviceArguments which could be empty.
     * 
     * This method does not return an IReturn. Instead, it accepts a callback specified by
     * listenerInfo and will call the listener back when the service is completed.
     * 
     * The listenerInfo contains enough info for the messenger to invoke a callback using
     * sendMessage(). Thus, the first element in the listenerInfo will be a String representing a
     * receiverService, and rest of elements representing serviceArguments.
     * 
     * The caller can also specify whether the callback wants the result of the service by placing
     * an integer prior the receiverService in the listenerInfo array. For example, a listenerInfo
     * with value like {1, "callbackId:callbackService", "123", null} means that the callback
     * service name is "callbackId:callbackService", the serviceArguments would be {"123", null}
     * where the second element is the placeholder of result as specified by 1.
     * 
     * @param listenerInfo
     *            callback's receiverService and serviceArguments
     * @param receiverService
     *            qualified service name
     * @param serviceArguments
     *            arguments for the service
     */
    void callServiceFor (Object[] listenerInfo, String receiverService, Object... serviceArguments);

    /**
     * Requests to invoke a service on an active object with parameters. This works the same as
     * callService() method does but it does not require a return value.
     * 
     * @param receiverService
     *            qualified service name
     * @param serviceArguments
     *            arguments for the service
     * 
     */
    void sendMessage (String receiverService, Object... serviceArguments);

    /**
     * Publishes a message with sending active object and service. The messenger will look up routes
     * registered previously using registerRoute() method to determine a list of target active
     * objects and services. If no target active objects are found, this method takes no action.
     * 
     * @param senderService
     *            sender's qualified service name
     * @param serviceArguments
     *            sender's message in form of arguments
     */
    void publishMessage (String senderService, Object... serviceArguments);

    /**
     * Creates a new active object with a unique name that is specified by parameter receiver, and a
     * list of services whose names are specified by parameter services.
     * 
     * And then binds an implementation object that is specified by parameter implObj to the active
     * object.
     * 
     * The active object created by this method is a single headcount. It means the single headcount
     * will execute service requests sequentially one at a time.
     * 
     * @param implObj
     *            implementation object
     * @param receiver
     *            receiver name
     * @param services
     *            a list of service names
     */
    void registerReceiver (Object implObj, String receiver, String... services);

    /**
     * Creates a new active object with a unique name that is specified by parameter receiver, and a
     * list of services whose names are specified by parameter services.
     * 
     * And then binds multiply implementation objects that are specified by parameter implObjs to
     * the active object.
     * 
     * The active object created by this method is a group of single headcounts. It means each
     * headcount will execute service requests sequentially one at a time. All together the active
     * object will execute service requests concurrently.
     * 
     * @param implObjs
     *            a list of implementation objects
     * @param receiver
     *            receiver name
     * @param services
     *            a list of service names
     */
    void registerReceiver (Object[] implObjs, String receiver, String... services);

    /**
     * Destroys an active object whose name is specified by parameter service, and releases its
     * implementation objects. All pending service requests will not be executed.
     * 
     * @param receiver
     *            receiver name
     */
    void deregisterReceiver (String receiver);

    /**
     * Adds a route
     * 
     * @param senderService
     *            sending qualified service name
     * @param receiverService
     *            receiving qualified service name
     */
    void registerRoute (String senderService, String receiverService);

    /**
     * Removes a route
     * 
     * @param senderService
     *            sending qualified service name
     * @param receiverService
     *            receiving qualified service name
     */
    void deregisterRoute (String senderService, String receiverService);

}
